
/********************************************
mawk_execute.c

libmawk changes (C) 2009-2010, Tibor 'Igor2' Palinkas;
based on mawk code coming with the below copyright:

copyright 1991-1996, Michael D. Brennan

This is a source file for mawk, an implementation of
the AWK programming language.

Mawk is distributed without warranty under the terms of
the GNU General Public License, version 2, 1991.
********************************************/

#include "mawk.h"
#include "code.h"
#include "memory.h"
#include "symtype.h"
#include "field.h"
#include "bi_funct.h"
#include "bi_vars.h"
#include "regexp.h"
#include "repl.h"
#include "fin.h"
#include "debug.h"
#include <string.h>
#include "num.h"

static int compare(mawk_state_t *, CELL *);
static int d_to_index(mawk_state_t *, mawk_num_t);


/*  The stack machine that mawk_executes the code */

#ifdef	DEBUG
static void eval_overflow(mawk_state_t * MAWK)
{
	mawk_overflow(MAWK, "eval stack", EVAL_STACK_SIZE);
}
#endif

/* holds info for array loops (on a stack) */
typedef struct aloop_state {
	struct aloop_state *link;
	CELL *var;										/* for(var in A) */
	STRING **base;
	STRING **ptr;
	STRING **limit;
} ALOOP_STATE;

/* clean up aloop stack on next, return, exit */
#define CLEAR_ALOOP_STACK() if(aloop_state){\
	    clear_aloop_stack(MAWK, aloop_state);\
	    aloop_state=(ALOOP_STATE*)0;}else

static void clear_aloop_stack(mawk_state_t *MAWK, ALOOP_STATE *top)
{
	ALOOP_STATE *q;

	do {
		while (top->ptr < top->limit) {
			free_STRING(*top->ptr);
			top->ptr++;
		}
		if (top->base < top->limit)
			zfree(MAWK, top->base, (top->limit - top->base) * sizeof(STRING *));
		q = top;
		top = q->link;
		ZFREE(q);
	} while (top);
}


/* Call awk function fbp with a_args already pushed on the stack. sp is
the stack pointer that is returned after the operation. */
CELL *mawk_call(mawk_state_t * MAWK, CELL * sp, FBLOCK * fbp, int a_args)
{
	CELL *cp;
	int t;
	CELL *nfp = sp - a_args + 1;	/* new fp for callee */
	CELL *local_p = sp + 1;				/* first local argument on stack */
	char *type_p;									/* pts to type of an argument */
	SYMTAB *sym;
	void *old_userdata;

	if (fbp->code == NULL) {
		sym = mawk_find(MAWK, fbp->name, 1);
		if (sym->type == ST_C_FUNCTION) {
			sp[1].type = ST_NONE;
			old_userdata = MAWK->userdata;
			MAWK->userdata = sym->stval.c_function.userdata;
			sp = sym->stval.c_function.callback(MAWK, sp, a_args) + 1;
			MAWK->userdata = old_userdata;
		}
		else {
			mawk_rt_error(MAWK, "unresolved function call %s\n", fbp->name);
		}
		return sp;
	}
	else {
		if (fbp->nargs)
			type_p = fbp->typev + a_args - 1;

		/* create space for locals */
		t = fbp->nargs - a_args;		/* t is number of locals */
		while (t > 0) {
			t--;
			sp++;
			type_p++;
			sp->type = C_NOINIT;
			if (*type_p == ST_LOCAL_ARRAY)
				sp->ptr = (PTR) new_ARRAY();
		}

		if (MAWK->debug_symbols)
			mawk_debug_callstack_push(MAWK, fbp);

		mawk_execute(MAWK, fbp->code, sp, nfp);

		if (MAWK->debug_symbols)
			mawk_debug_callstack_pop(MAWK);

	}

	/* cleanup the callee's arguments */
	/* putting return value at top of eval stack */
	if (sp >= nfp) {
		cp = sp + 1;								/* cp -> the function return */

		do {
			if (*type_p == ST_LOCAL_ARRAY) {
				if (sp >= local_p) {
					mawk_array_clear(MAWK, sp->ptr);
					ZFREE((ARRAY) sp->ptr);
				}
			}
			else
				cell_destroy(sp);
			type_p--;
			sp--;
		}
		while (sp >= nfp);

		mawk_cellcpy(MAWK, ++sp, cp);
		cell_destroy(cp);
	}
	else
		sp++;												/* no arguments passed */
	return sp;
}

void mawk_execute(mawk_state_t *MAWK, register INST *cdp, register CELL *sp, CELL *fp)
{
	/* cdp -> code ptr, start execution here */
	/* sp -> eval_stack pointer */
	/* fp -> frame ptr into eval_stack for user defined functions */

	/* some useful temporaries */
	CELL *cp;
	int t;

	/* save state for array loops via a stack */
	ALOOP_STATE *aloop_state = (ALOOP_STATE *) 0;

	/* for moving the eval stack on deep recursion */
	CELL *old_stack_base;
	CELL *old_sp;

#ifdef	DEBUG
	CELL *entry_sp = sp;
#endif


	if (fp) {
		/* we are a function call, check for deep recursion */
		if (sp > MAWK->stack_danger) {	/* change stacks */
			old_stack_base = MAWK->stack_base;
			old_sp = sp;
			MAWK->stack_base = (CELL *) zmalloc(MAWK, sizeof(CELL) * EVAL_STACK_SIZE);
			MAWK->stack_danger = MAWK->stack_base + DANGER;
			sp = MAWK->stack_base;
			/* waste 1 slot for ANSI, actually large model msdos breaks in
			   RET if we don't */
#ifdef	DEBUG
			entry_sp = sp;
#endif
		}
		else
			old_stack_base = (CELL *) 0;
	}

	while (1)
		switch (cdp++->op) {

/* HALT only used by the disassemble now ; this remains
   so compilers don't offset the jump table */
		case _HALT:

		case _STOP:								/* only for range patterns */
#ifdef	DEBUG
			if (sp != entry_sp + 1)
				mawk_bozo(MAWK, "stop0");
#endif
			return;

		case _PUSHC:
			inc_sp();
			mawk_cellcpy(MAWK, sp, cdp++->ptr);
			break;

		case _PUSHD:
			inc_sp();
			sp->type = C_NUM;
			sp->dval = *(mawk_num_t *) cdp++->ptr;
			break;

		case _PUSHS:
			inc_sp();
			sp->type = C_STRING;
			sp->ptr = cdp++->ptr;
			string(sp)->ref_cnt++;
			break;

		case F_PUSHA:
			cp = (CELL *) cdp->ptr;
			if (cp != MAWK->field) {
				if (MAWK->nf < 0)
					mawk_split_field0(MAWK);

				if (!(cp >= NF && cp <= LAST_PFIELD)) {
					/* its a real field $1, $2 ...
					   If its greater than $NF, we have to
					   make sure its set to ""  so that
					   (++|--) and g?sub() work right
					 */
					t = mawk_field_addr_to_index(MAWK, cp);
					if (t > MAWK->nf) {
						cell_destroy(cp);
						cp->type = C_STRING;
						cp->ptr = (PTR) & (MAWK->null_str);
						MAWK->null_str.ref_cnt++;
					}
				}
			}
			/* fall thru */

		case _PUSHA:
		case A_PUSHA:
			inc_sp();
			sp->ptr = cdp++->ptr;
			break;

		case _PUSHI:
			/* put contents of next address on stack */
			inc_sp();
			mawk_cellcpy(MAWK, sp, cdp++->ptr);
			break;

		case L_PUSHI:
			/* put the contents of a local var on stack,
			   cdp->op holds the offset from the frame pointer */
			inc_sp();
			mawk_cellcpy(MAWK, sp, fp + cdp++->op);
			break;

		case L_PUSHA:
			/* put a local address on eval stack */
			inc_sp();
			sp->ptr = (PTR) (fp + cdp++->op);
			break;


		case F_PUSHI:

			/* push contents of $i
			   cdp[0] holds & $i , cdp[1] holds i */

			inc_sp();
			if (MAWK->nf < 0)
				mawk_split_field0(MAWK);
			cp = (CELL *) cdp->ptr;
			t = (cdp + 1)->op;
			cdp += 2;

			if (t <= MAWK->nf)
				mawk_cellcpy(MAWK, sp, cp);
			else {										/* an unset field */

				sp->type = C_STRING;
				sp->ptr = (PTR) & (MAWK->null_str);
				MAWK->null_str.ref_cnt++;
			}
			break;

		case NF_PUSHI:

			inc_sp();
			if (MAWK->nf < 0)
				mawk_split_field0(MAWK);
			mawk_cellcpy(MAWK, sp, NF);
			break;

		case FE_PUSHA:

			if (sp->type != C_NUM)
				mawk_cast1_to_d(MAWK, sp);

			t = d_to_index(MAWK, sp->dval);
			if (t && MAWK->nf < 0)
				mawk_split_field0(MAWK);
			sp->ptr = (PTR) field_ptr(t);
			if (t > MAWK->nf) {
				/* make sure its set to "" */
				cp = (CELL *) sp->ptr;
				cell_destroy(cp);
				cp->type = C_STRING;
				cp->ptr = (PTR) & (MAWK->null_str);
				MAWK->null_str.ref_cnt++;
			}
			break;

		case FE_PUSHI:

			if (sp->type != C_NUM)
				mawk_cast1_to_d(MAWK, sp);

			t = d_to_index(MAWK, sp->dval);

			if (MAWK->nf < 0)
				mawk_split_field0(MAWK);
			if (t <= MAWK->nf)
				mawk_cellcpy(MAWK, sp, field_ptr(t));
			else {
				sp->type = C_STRING;
				sp->ptr = (PTR) & MAWK->null_str;
				MAWK->null_str.ref_cnt++;
			}
			break;


		case AE_PUSHA:
			/* top of stack has an expr, cdp->ptr points at an
			   array, replace the expr with the cell address inside
			   the array */

			cp = mawk_array_find(MAWK, (ARRAY) cdp++->ptr, sp, CREATE);
			cell_destroy(sp);
			sp->ptr = (PTR) cp;
			break;

		case AE_PUSHI:
			/* top of stack has an expr, cdp->ptr points at an
			   array, replace the expr with the contents of the
			   cell inside the array */

			cp = mawk_array_find(MAWK, (ARRAY) cdp++->ptr, sp, CREATE);
			cell_destroy(sp);
			mawk_cellcpy(MAWK, sp, cp);
			break;

		case LAE_PUSHI:
			/*  sp[0] is an expression
			   cdp->op is offset from frame pointer of a CELL which
			   has an ARRAY in the ptr field, replace expr
			   with  array[expr]
			 */
			cp = mawk_array_find(MAWK, (ARRAY) fp[cdp++->op].ptr, sp, CREATE);
			cell_destroy(sp);
			mawk_cellcpy(MAWK, sp, cp);
			break;

		case LAE_PUSHA:
			/*  sp[0] is an expression
			   cdp->op is offset from frame pointer of a CELL which
			   has an ARRAY in the ptr field, replace expr
			   with  & array[expr]
			 */
			cp = mawk_array_find(MAWK, (ARRAY) fp[cdp++->op].ptr, sp, CREATE);
			cell_destroy(sp);
			sp->ptr = (PTR) cp;
			break;

		case LA_PUSHA:
			/*  cdp->op is offset from frame pointer of a CELL which
			   has an ARRAY in the ptr field. Push this ARRAY
			   on the eval stack
			 */
			inc_sp();
			sp->ptr = fp[cdp++->op].ptr;
			break;

		case SET_ALOOP:
			{
				ALOOP_STATE *ap = ZMALLOC(ALOOP_STATE);
				unsigned vector_size;

				ap->var = (CELL *) sp[-1].ptr;
				ap->base = ap->ptr = mawk_array_loop_vector(MAWK, (ARRAY) sp->ptr, &vector_size);
				ap->limit = ap->base + vector_size;
				sp -= 2;

				/* push onto aloop stack */
				ap->link = aloop_state;
				aloop_state = ap;
				cdp += cdp->op;
			}
			break;

		case ALOOP:
			{
				ALOOP_STATE *ap = aloop_state;
				if (ap->ptr < ap->limit) {
					cell_destroy(ap->var);
					ap->var->type = C_STRING;
					ap->var->ptr = (PTR) * ap->ptr++;
					cdp += cdp->op;
				}
				else
					cdp++;
			}
			break;

		case POP_AL:
			{
				/* finish up an array loop */
				ALOOP_STATE *ap = aloop_state;
				aloop_state = ap->link;
				while (ap->ptr < ap->limit) {
					free_STRING(*ap->ptr);
					ap->ptr++;
				}
				if (ap->base < ap->limit)
					zfree(MAWK, ap->base, (ap->limit - ap->base) * sizeof(STRING *));
				ZFREE(ap);
			}
			break;

		case _POP:
			cell_destroy(sp);
			sp--;
			break;

		case _ASSIGN:
			/* top of stack has an expr, next down is an
			   address, put the expression in *address and
			   replace the address with the expression */

			/* don't propagate type C_MBSTRN */
			if (sp->type == C_MBSTRN)
				mawk_check_strnum(MAWK, sp);
			sp--;
			cell_destroy(((CELL *) sp->ptr));
			mawk_cellcpy(MAWK, sp, mawk_cellcpy(MAWK, sp->ptr, sp + 1));
			cell_destroy(sp + 1);
			break;

		case F_ASSIGN:
			/* assign to a field  */
			if (sp->type == C_MBSTRN)
				mawk_check_strnum(MAWK, sp);
			sp--;
			mawk_field_assign(MAWK, (CELL *) sp->ptr, sp + 1);
			cell_destroy(sp + 1);
			mawk_cellcpy(MAWK, sp, (CELL *) sp->ptr);
			break;

		case _ADD_ASG:
			if (sp->type != C_NUM)
				mawk_cast1_to_d(MAWK, sp);
			cp = (CELL *) (sp - 1)->ptr;
			if (cp->type != C_NUM)
				mawk_cast1_to_d(MAWK, cp);

#if SW_FP_CHECK									/* specific to V7 and XNX23A */
			clrerr();
#endif
			cp->dval += sp--->dval;
#if SW_FP_CHECK
			fpcheck();
#endif
			sp->type = C_NUM;
			sp->dval = cp->dval;
			break;

		case _SUB_ASG:
			if (sp->type != C_NUM)
				mawk_cast1_to_d(MAWK, sp);
			cp = (CELL *) (sp - 1)->ptr;
			if (cp->type != C_NUM)
				mawk_cast1_to_d(MAWK, cp);
#if SW_FP_CHECK
			clrerr();
#endif
			cp->dval -= sp--->dval;
#if SW_FP_CHECK
			fpcheck();
#endif
			sp->type = C_NUM;
			sp->dval = cp->dval;
			break;

		case _MUL_ASG:
			if (sp->type != C_NUM)
				mawk_cast1_to_d(MAWK, sp);
			cp = (CELL *) (sp - 1)->ptr;
			if (cp->type != C_NUM)
				mawk_cast1_to_d(MAWK, cp);
#if SW_FP_CHECK
			clrerr();
#endif
			cp->dval *= sp--->dval;
#if SW_FP_CHECK
			fpcheck();
#endif
			sp->type = C_NUM;
			sp->dval = cp->dval;
			break;

		case _DIV_ASG:
			if (sp->type != C_NUM)
				mawk_cast1_to_d(MAWK, sp);
			cp = (CELL *) (sp - 1)->ptr;
			if (cp->type != C_NUM)
				mawk_cast1_to_d(MAWK, cp);

#if  NOINFO_SIGFPE
			CHECK_DIVZERO(sp->dval);
#endif

#ifdef MAWK_NO_FLOAT
			{
				mawk_num_t d;
				d = sp--->dval;
				if (d != NUM_ZERO)
					cp->dval /= d;
				else
					cp->dval = NUM_INF;
			}
#else
#	if SW_FP_CHECK
			clrerr();
#	endif
			cp->dval /= sp--->dval;
#	if SW_FP_CHECK
			fpcheck();
#	endif
#endif
			sp->type = C_NUM;
			sp->dval = cp->dval;
			break;

		case _MOD_ASG:
			if (sp->type != C_NUM)
				mawk_cast1_to_d(MAWK, sp);
			cp = (CELL *) (sp - 1)->ptr;
			if (cp->type != C_NUM)
				mawk_cast1_to_d(MAWK, cp);

#ifdef MAWK_NO_FLOAT
			{
				int d;
				d = sp--->dval;
				if (d != NUM_ZERO)
					cp->dval %= d;
				else
					cp->dval = NUM_NAN;
			}
#else
#	if  NOINFO_SIGFPE
			CHECK_DIVZERO(sp->dval);
#	endif

			cp->dval = fmod(cp->dval, sp--->dval);
#endif
			sp->type = C_NUM;
			sp->dval = cp->dval;
			break;

		case _POW_ASG:
			if (sp->type != C_NUM)
				mawk_cast1_to_d(MAWK, sp);
			cp = (CELL *) (sp - 1)->ptr;
			if (cp->type != C_NUM)
				mawk_cast1_to_d(MAWK, cp);
			cp->dval = mawk_num_pow(cp->dval, sp--->dval);
			sp->type = C_NUM;
			sp->dval = cp->dval;
			break;

			/* will anyone ever use these ? */

		case F_ADD_ASG:
			if (sp->type != C_NUM)
				mawk_cast1_to_d(MAWK, sp);
			cp = (CELL *) (sp - 1)->ptr;
			mawk_cast1_to_d(MAWK, mawk_cellcpy(MAWK, &MAWK->tc, cp));
#if SW_FP_CHECK
			clrerr();
#endif
			MAWK->tc.dval += sp--->dval;
#if SW_FP_CHECK
			fpcheck();
#endif
			sp->type = C_NUM;
			sp->dval = MAWK->tc.dval;
			mawk_field_assign(MAWK, cp, &MAWK->tc);
			break;

		case F_SUB_ASG:
			if (sp->type != C_NUM)
				mawk_cast1_to_d(MAWK, sp);
			cp = (CELL *) (sp - 1)->ptr;
			mawk_cast1_to_d(MAWK, mawk_cellcpy(MAWK, &MAWK->tc, cp));
#if SW_FP_CHECK
			clrerr();
#endif
			MAWK->tc.dval -= sp--->dval;
#if SW_FP_CHECK
			fpcheck();
#endif
			sp->type = C_NUM;
			sp->dval = MAWK->tc.dval;
			mawk_field_assign(MAWK, cp, &MAWK->tc);
			break;

		case F_MUL_ASG:
			if (sp->type != C_NUM)
				mawk_cast1_to_d(MAWK, sp);
			cp = (CELL *) (sp - 1)->ptr;
			mawk_cast1_to_d(MAWK, mawk_cellcpy(MAWK, &MAWK->tc, cp));
#if SW_FP_CHECK
			clrerr();
#endif
			MAWK->tc.dval *= sp--->dval;
#if SW_FP_CHECK
			fpcheck();
#endif
			sp->type = C_NUM;
			sp->dval = MAWK->tc.dval;
			mawk_field_assign(MAWK, cp, &MAWK->tc);
			break;

		case F_DIV_ASG:
			if (sp->type != C_NUM)
				mawk_cast1_to_d(MAWK, sp);
			cp = (CELL *) (sp - 1)->ptr;
			mawk_cast1_to_d(MAWK, mawk_cellcpy(MAWK, &MAWK->tc, cp));

#if  NOINFO_SIGFPE
			CHECK_DIVZERO(sp->dval);
#endif

#ifdef MAWK_NO_FLOAT
		{
			mawk_num_t d;
			d = sp--->dval;
			if (d != NUM_ZERO)
				MAWK->tc.dval /= d;
			else
				MAWK->tc.dval = NUM_INF;
		}
#else
#	if SW_FP_CHECK
			clrerr();
#	endif
			MAWK->tc.dval /= sp--->dval;
#	if SW_FP_CHECK
			fpcheck();
#	endif
#endif
			sp->type = C_NUM;
			sp->dval = MAWK->tc.dval;
			mawk_field_assign(MAWK, cp, &MAWK->tc);
			break;

		case F_MOD_ASG:
			if (sp->type != C_NUM)
				mawk_cast1_to_d(MAWK, sp);
			cp = (CELL *) (sp - 1)->ptr;
			mawk_cast1_to_d(MAWK, mawk_cellcpy(MAWK, &MAWK->tc, cp));

#ifdef MAWK_NO_FLOAT
			{
				int d;
				d = sp--->dval;
				if (d != NUM_ZERO)
					MAWK->tc.dval %= d;
				else
					MAWK->tc.dval = NUM_NAN;
			}
#else
#	if  NOINFO_SIGFPE
			CHECK_DIVZERO(sp->dval);
#	endif

			MAWK->tc.dval = fmod(MAWK->tc.dval, sp--->dval);
#endif
			sp->type = C_NUM;
			sp->dval = MAWK->tc.dval;
			mawk_field_assign(MAWK, cp, &MAWK->tc);
			break;

		case F_POW_ASG:
			if (sp->type != C_NUM)
				mawk_cast1_to_d(MAWK, sp);
			cp = (CELL *) (sp - 1)->ptr;
			mawk_cast1_to_d(MAWK, mawk_cellcpy(MAWK, &MAWK->tc, cp));
			MAWK->tc.dval = mawk_num_pow(MAWK->tc.dval, sp--->dval);
			sp->type = C_NUM;
			sp->dval = MAWK->tc.dval;
			mawk_field_assign(MAWK, cp, &MAWK->tc);
			break;

		case _ADD:
			sp--;
			if (TEST2(sp) != TWO_NUMS)
				mawk_cast2_to_d(MAWK, sp);
#if SW_FP_CHECK
			clrerr();
#endif
			sp[0].dval += sp[1].dval;
#if SW_FP_CHECK
			fpcheck();
#endif
			break;

		case _SUB:
			sp--;
			if (TEST2(sp) != TWO_NUMS)
				mawk_cast2_to_d(MAWK, sp);
#if SW_FP_CHECK
			clrerr();
#endif
			sp[0].dval -= sp[1].dval;
#if SW_FP_CHECK
			fpcheck();
#endif
			break;

		case _MUL:
			sp--;
			if (TEST2(sp) != TWO_NUMS)
				mawk_cast2_to_d(MAWK, sp);
#if SW_FP_CHECK
			clrerr();
#endif
			sp[0].dval *= sp[1].dval;
#if SW_FP_CHECK
			fpcheck();
#endif
			break;

		case _DIV:
			sp--;
			if (TEST2(sp) != TWO_NUMS)
				mawk_cast2_to_d(MAWK, sp);

#if  NOINFO_SIGFPE
			CHECK_DIVZERO(sp[1].dval);
#endif

#ifdef MAWK_NO_FLOAT
			if (sp[1].dval != 0)
				sp[0].dval /= sp[1].dval;
			else
				sp[0].dval = NUM_INF;
#else
#	if SW_FP_CHECK
			clrerr();
#	endif
			sp[0].dval /= sp[1].dval;
#	if SW_FP_CHECK
			fpcheck();
#	endif
#endif
			break;

		case _MOD:
			sp--;
			if (TEST2(sp) != TWO_NUMS)
				mawk_cast2_to_d(MAWK, sp);

#ifdef MAWK_NO_FLOAT
			{
				int d;
				d = sp[1].dval;
				if (d != NUM_ZERO)
					sp[0].dval %= d;
				else
					sp[0].dval = NUM_NAN;
			}
#else

#	if  NOINFO_SIGFPE
			CHECK_DIVZERO(sp[1].dval);
#	endif
			sp[0].dval = fmod(sp[0].dval, sp[1].dval);
#endif
			break;

		case _POW:
			sp--;
			if (TEST2(sp) != TWO_NUMS)
				mawk_cast2_to_d(MAWK, sp);
			sp[0].dval = mawk_num_pow(sp[0].dval, sp[1].dval);
			break;

		case _NOT:
			/* evaluates to 0.0 or 1.0 */
		reswitch_1:
			switch (sp->type) {
			case C_NOINIT:
				sp->dval = NUM_ONE;
				break;
			case C_NUM:
				sp->dval = sp->dval != NUM_ZERO ? NUM_ZERO : NUM_ONE;
				break;
			case C_STRING:
				sp->dval = string(sp)->len ? NUM_ZERO : NUM_ONE;
				free_STRING(string(sp));
				break;
			case C_STRNUM:						/* mawk_test as a number */
				sp->dval = sp->dval != NUM_ZERO ? NUM_ZERO : NUM_ONE;
				free_STRING(string(sp));
				break;
			case C_MBSTRN:
				mawk_check_strnum(MAWK, sp);
				goto reswitch_1;
			default:
				mawk_bozo(MAWK, "bad type on eval stack");
			}
			sp->type = C_NUM;
			break;

		case _TEST:
			/* evaluates to 0.0 or 1.0 */
		reswitch_2:
			switch (sp->type) {
			case C_NOINIT:
				sp->dval = NUM_ZERO;
				break;
			case C_NUM:
				sp->dval = sp->dval != NUM_ZERO ? NUM_ONE : NUM_ZERO;
				break;
			case C_STRING:
				sp->dval = string(sp)->len ? NUM_ONE : NUM_ZERO;
				free_STRING(string(sp));
				break;
			case C_STRNUM:						/* mawk_test as a number */
				sp->dval = sp->dval != NUM_ZERO ? NUM_ONE : NUM_ZERO;
				free_STRING(string(sp));
				break;
			case C_MBSTRN:
				mawk_check_strnum(MAWK, sp);
				goto reswitch_2;
			default:
				mawk_bozo(MAWK, "bad type on eval stack");
			}
			sp->type = C_NUM;
			break;

		case _UMINUS:
			if (sp->type != C_NUM)
				mawk_cast1_to_d(MAWK, sp);
			sp->dval = -sp->dval;
			break;

		case _UPLUS:
			if (sp->type != C_NUM)
				mawk_cast1_to_d(MAWK, sp);
			break;

		case _CAT:
			{
				unsigned len1, len2;
				char *str1, *str2;
				STRING *b;

				sp--;
				if (TEST2(sp) != TWO_STRINGS)
					mawk_cast2_to_s(MAWK, sp);
				str1 = string(sp)->str;
				len1 = string(sp)->len;
				str2 = string(sp + 1)->str;
				len2 = string(sp + 1)->len;

				b = mawk_new_STRING0(MAWK, len1 + len2);
				memcpy(b->str, str1, len1);
				memcpy(b->str + len1, str2, len2);
				free_STRING(string(sp));
				free_STRING(string(sp + 1));

				sp->ptr = (PTR) b;
				break;
			}

		case _PUSHINT:
			inc_sp();
			sp->type = cdp++->op;
			break;

		case _BUILTIN:
		case _PRINT:
			sp = (*(PF_CP) cdp++->ptr) (MAWK, sp);
			break;

		case _POST_INC:
			cp = (CELL *) sp->ptr;
			if (cp->type != C_NUM)
				mawk_cast1_to_d(MAWK, cp);
			sp->type = C_NUM;
			sp->dval = cp->dval;
			cp->dval += NUM_ONE;
			break;

		case _POST_DEC:
			cp = (CELL *) sp->ptr;
			if (cp->type != C_NUM)
				mawk_cast1_to_d(MAWK, cp);
			sp->type = C_NUM;
			sp->dval = cp->dval;
			cp->dval -= NUM_ONE;
			break;

		case _PRE_INC:
			cp = (CELL *) sp->ptr;
			if (cp->type != C_NUM)
				mawk_cast1_to_d(MAWK, cp);
			sp->dval = cp->dval += NUM_ONE;
			sp->type = C_NUM;
			break;

		case _PRE_DEC:
			cp = (CELL *) sp->ptr;
			if (cp->type != C_NUM)
				mawk_cast1_to_d(MAWK, cp);
			sp->dval = cp->dval -= NUM_ONE;
			sp->type = C_NUM;
			break;


		case F_POST_INC:
			cp = (CELL *) sp->ptr;
			mawk_cellcpy(MAWK, &MAWK->tc, cp);
			mawk_cast1_to_d(MAWK, &MAWK->tc);
			sp->type = C_NUM;
			sp->dval = MAWK->tc.dval;
			MAWK->tc.dval += NUM_ONE;
			mawk_field_assign(MAWK, cp, &MAWK->tc);
			break;

		case F_POST_DEC:
			cp = (CELL *) sp->ptr;
			mawk_cellcpy(MAWK, &MAWK->tc, cp);
			mawk_cast1_to_d(MAWK, &MAWK->tc);
			sp->type = C_NUM;
			sp->dval = MAWK->tc.dval;
			MAWK->tc.dval -= NUM_ONE;
			mawk_field_assign(MAWK, cp, &MAWK->tc);
			break;

		case F_PRE_INC:
			cp = (CELL *) sp->ptr;
			mawk_cast1_to_d(MAWK, mawk_cellcpy(MAWK, sp, cp));
			sp->dval += NUM_ONE;
			mawk_field_assign(MAWK, cp, sp);
			break;

		case F_PRE_DEC:
			cp = (CELL *) sp->ptr;
			mawk_cast1_to_d(MAWK, mawk_cellcpy(MAWK, sp, cp));
			sp->dval -= NUM_ONE;
			mawk_field_assign(MAWK, cp, sp);
			break;

		case _JMP:
			cdp += cdp->op;
			break;

		case _JNZ:
			/* jmp if top of stack is non-zero and pop stack */
			if (mawk_test(MAWK, sp))
				cdp += cdp->op;
			else
				cdp++;
			cell_destroy(sp);
			sp--;
			break;

		case _JZ:
			/* jmp if top of stack is zero and pop stack */
			if (!mawk_test(MAWK, sp))
				cdp += cdp->op;
			else
				cdp++;
			cell_destroy(sp);
			sp--;
			break;

		case _LJZ:
			/* special jump for logical and */
			/* this is always preceded by _TEST */
			if (sp->dval == NUM_ZERO) {
				/* take jump, but don't pop stack */
				cdp += cdp->op;
			}
			else {
				/* pop and don't jump */
				sp--;
				cdp++;
			}
			break;

		case _LJNZ:
			/* special jump for logical or */
			/* this is always preceded by _TEST */
			if (sp->dval != NUM_ZERO) {
				/* take jump, but don't pop stack */
				cdp += cdp->op;
			}
			else {
				/* pop and don't jump */
				sp--;
				cdp++;
			}
			break;

			/*  the relation operations */
			/*  compare() makes sure string ref counts are OK */
		case _EQ:
			t = compare(MAWK, --sp);
			sp->type = C_NUM;
			sp->dval = t == 0 ? NUM_ONE : NUM_ZERO;
			break;

		case _NEQ:
			t = compare(MAWK, --sp);
			sp->type = C_NUM;
			sp->dval = t ? NUM_ONE : NUM_ZERO;
			break;

		case _LT:
			t = compare(MAWK, --sp);
			sp->type = C_NUM;
			sp->dval = t < 0 ? NUM_ONE : NUM_ZERO;
			break;

		case _LTE:
			t = compare(MAWK, --sp);
			sp->type = C_NUM;
			sp->dval = t <= 0 ? NUM_ONE : NUM_ZERO;
			break;

		case _GT:
			t = compare(MAWK, --sp);
			sp->type = C_NUM;
			sp->dval = t > 0 ? NUM_ONE : NUM_ZERO;
			break;

		case _GTE:
			t = compare(MAWK, --sp);
			sp->type = C_NUM;
			sp->dval = t >= 0 ? NUM_ONE : NUM_ZERO;
			break;

		case _MATCH0:
			/* does $0 match, the RE at cdp? */

			inc_sp();
			if (MAWK->field->type >= C_STRING) {
				sp->type = C_NUM;
				sp->dval = mawk_REtest(string(MAWK->field)->str, cdp++->ptr)
					? NUM_ONE : NUM_ZERO;

				break /* the case */ ;
			}
			else {
				mawk_cellcpy(MAWK, sp, MAWK->field);
				/* and FALL THRU */
			}

		case _MATCH1:
			/* does expr at sp[0] match RE at cdp */
			if (sp->type < C_STRING)
				mawk_cast1_to_s(MAWK, sp);
			t = mawk_REtest(string(sp)->str, cdp++->ptr);
			free_STRING(string(sp));
			sp->type = C_NUM;
			sp->dval = t ? NUM_ONE : NUM_ZERO;
			break;


		case _MATCH2:
			/* does sp[-1] match sp[0] as re */
			mawk_cast_to_RE(MAWK, sp);

			if ((--sp)->type < C_STRING)
				mawk_cast1_to_s(MAWK, sp);
			t = mawk_REtest(string(sp)->str, (sp + 1)->ptr);

			free_STRING(string(sp));
			sp->type = C_NUM;
			sp->dval = t ? NUM_ONE : NUM_ZERO;
			break;

		case A_TEST:
			/* entry :  sp[0].ptr-> an array
			   sp[-1]  is an expression

			   we compute (expression in array)  */
			sp--;
			cp = mawk_array_find(MAWK, (sp + 1)->ptr, sp, NO_CREATE);
			cell_destroy(sp);
			sp->type = C_NUM;
			sp->dval = (cp != (CELL *) 0) ? NUM_ONE : NUM_ZERO;
			break;

		case A_DEL:
			/* sp[0].ptr ->  array
			   sp[-1] is an expr
			   delete  array[expr]  */

			mawk_array_delete(MAWK, sp->ptr, sp - 1);
			cell_destroy(sp - 1);
			sp -= 2;
			break;

		case DEL_A:
			/* free all the array at once */
			mawk_array_clear(MAWK, sp->ptr);
			sp--;
			break;

			/* form a multiple array index */
		case A_CAT:
			sp = mawk_array_cat(MAWK, sp, cdp++->op);
			break;

		case _EXIT:
			if (sp->type != C_NUM)
				mawk_cast1_to_d(MAWK, sp);
			MAWK->exit_code = d_to_i(sp->dval);
			sp--;
			/* fall thru */

		case _EXIT0:

			if (!MAWK->end_start)
				mawk_exit(MAWK, MAWK->exit_code);


			cdp = MAWK->end_start;
			MAWK->end_start = (INST *) 0;	/* makes sure next exit exits */

			if (MAWK->begin_start)
				zfree(MAWK, MAWK->begin_start, MAWK->begin_size);
			if (MAWK->main_start)
				zfree(MAWK, MAWK->main_start, MAWK->main_size);
			sp = MAWK->eval_stack - 1;	/* might be in user function */
			CLEAR_ALOOP_STACK();			/* ditto */
			break;

		case _JMAIN:								/* go from BEGIN code to MAIN code */
/*			zfree(MAWK, MAWK->begin_start, MAWK->begin_size); (causes double free!) */
			MAWK->begin_start = (INST *) 0;
			cdp = MAWK->main_start;
			if (MAWK->separate_begin)
				return;
			break;

		case _OMAIN:
			if (!MAWK->main_fin)
				mawk_open_main(MAWK);
			if (!MAWK->main_fin)
				return;
			MAWK->restart_label = cdp;
			cdp = MAWK->next_label;
			break;

		case _NEXT:
			/* next might be inside an aloop -- clear stack */
			CLEAR_ALOOP_STACK();
			cdp = MAWK->next_label;
			break;

		case OL_GL:
			{
				char *p;
				unsigned len;

				p = mawk_FINgets(MAWK, MAWK->main_fin, &len);
				if (p == (void *) mawk_FIN_nomore)
					return;
				if (!p) {
					if (!MAWK->end_start)
						mawk_exit(MAWK, 0);

					cdp = MAWK->end_start;
/*					zfree(MAWK, MAWK->main_start, MAWK->main_size); (causes double free) */
					MAWK->main_start = MAWK->end_start = (INST *) 0;
				}
				else {
					mawk_set_field0(MAWK, p, len);
					cdp = MAWK->restart_label;
					MAWK->rt_nr++;
					MAWK->rt_fnr++;
				}
			}
			break;

			/* two kinds of OL_GL is a historical stupidity from working on
			   a machine with very slow floating point emulation */
		case OL_GL_NR:
			{
				char *p;
				unsigned len;
				p = mawk_FINgets(MAWK, MAWK->main_fin, &len);
				if (p == (void *) mawk_FIN_nomore)
					return;
				if (!p) {
					if (!MAWK->end_start)
						mawk_exit(MAWK, 0);

					cdp = MAWK->end_start;
					zfree(MAWK, MAWK->main_start, MAWK->main_size);
					MAWK->main_start = MAWK->end_start = (INST *) 0;
				}
				else {
					mawk_set_field0(MAWK, p, len);
					cdp = MAWK->restart_label;

					if (TEST2(NR) != TWO_NUMS) {
						mawk_cast2_to_d(MAWK, NR);
					}

					NR->dval += NUM_ONE;
					MAWK->rt_nr++;
					FNR->dval += NUM_ONE;
					MAWK->rt_fnr++;
				}
			}
			break;

		case _RANGE:
/* mawk_test a range pattern:  pat1, pat2 { action }
   entry :
       cdp[0].op -- a flag, mawk_test pat1 if on else pat2
       cdp[1].op -- offset of pat2 code from cdp
       cdp[2].op -- offset of action code from cdp
       cdp[3].op -- offset of code after the action from cdp
       cdp[4] -- start of pat1 code
*/

#define FLAG	cdp[0].op
#define PAT2	cdp[1].op
#define ACTION	  cdp[2].op
#define FOLLOW	  cdp[3].op
#define PAT1	  4

			if (FLAG) {								/* mawk_test again pat1 */
				mawk_execute(MAWK, cdp + PAT1, sp, fp);
				t = mawk_test(MAWK, sp + 1);
				cell_destroy(sp + 1);
				if (t)
					FLAG = 0;
				else {
					cdp += FOLLOW;
					break;								/* break the switch */
				}
			}

			/* mawk_test against pat2 and then perform the action */
			mawk_execute(MAWK, cdp + PAT2, sp, fp);
			FLAG = mawk_test(MAWK, sp + 1);
			cell_destroy(sp + 1);
			cdp += ACTION;
			break;

/* function calls  */

		case _RET0:
			inc_sp();
			sp->type = C_NOINIT;
			/* fall thru */

		case _RET:

#ifdef	DEBUG
			if (sp != entry_sp + 1)
				mawk_bozo(MAWK, "ret");
#endif
			if (old_stack_base) {			/* reset stack */
				/* move the return value */
				mawk_cellcpy(MAWK, old_sp + 1, sp);
				cell_destroy(sp);
				zfree(MAWK, MAWK->stack_base, sizeof(CELL) * EVAL_STACK_SIZE);
				MAWK->stack_base = old_stack_base;
				MAWK->stack_danger = old_stack_base + DANGER;
			}

			/* return might be inside an aloop -- clear stack */
			CLEAR_ALOOP_STACK();

			return;

		case _CALL:
			{
				/*  cdp[0] holds ptr to "function block"
				   cdp[1] holds number of input arguments
				 */
				FBLOCK *fbp = (FBLOCK *) cdp++->ptr;
				int a_args = cdp++->op;	/* actual number of args */

				sp = mawk_call(MAWK, sp, fbp, a_args);
			}
			break;
		case LOCATION:
			mawk_location_change(MAWK, cdp->op);
			cdp++;
			break;
		default:
			mawk_bozo(MAWK, "bad opcode");
		}
}


/*
  return 0 if a numeric is zero else return non-zero
  return 0 if a string is "" else return non-zero
*/
int mawk_test(mawk_state_t *MAWK, register CELL *cp)
{
reswitch:

	switch (cp->type) {
	case C_NOINIT:
		return 0;
	case C_STRNUM:								/* mawk_test as a number */
	case C_NUM:
		return cp->dval != NUM_ZERO;
	case C_STRING:
		return string(cp)->len;
	case C_MBSTRN:
		mawk_check_strnum(MAWK, cp);
		goto reswitch;
	default:
		mawk_bozo(MAWK, "bad cell type in call to mawk_test");
	}
	return 0;											/*can't get here: shutup */
}

/* compare cells at cp and cp+1 and
   frees STRINGs at those cells
*/
static int compare(mawk_state_t *MAWK, register CELL *cp)
{
	int k;

reswitch:

	switch (TEST2(cp)) {
	case TWO_NOINITS:
		return 0;

	case TWO_NUMS:
	two_d:
		return cp->dval > (cp + 1)->dval ? 1 : cp->dval < (cp + 1)->dval ? -1 : 0;

	case TWO_STRINGS:
	case STRING_AND_STRNUM:
	two_s:
		k = strcmp(string(cp)->str, string(cp + 1)->str);
		free_STRING(string(cp));
		free_STRING(string(cp + 1));
		return k;

	case NOINIT_AND_NUM:
	case NOINIT_AND_STRNUM:
	case NUM_AND_STRNUM:
	case TWO_STRNUMS:
		mawk_cast2_to_d(MAWK, cp);
		goto two_d;
	case NOINIT_AND_STRING:
	case NUM_AND_STRING:
		mawk_cast2_to_s(MAWK, cp);
		goto two_s;
	case TWO_MBSTRNS:
		mawk_check_strnum(MAWK, cp);
		mawk_check_strnum(MAWK, cp + 1);
		goto reswitch;

	case NOINIT_AND_MBSTRN:
	case NUM_AND_MBSTRN:
	case STRING_AND_MBSTRN:
	case STRNUM_AND_MBSTRN:
		mawk_check_strnum(MAWK, cp->type == C_MBSTRN ? cp : cp + 1);
		goto reswitch;

	default:											/* there are no default cases */
		mawk_bozo(MAWK, "bad cell type passed to compare");
	}
	return 0;											/* shut up */
}

/* does not assume target was a cell, if so
   then caller should have made a previous
   call to cell_destroy	 */
CELL *mawk_cellcpy(mawk_state_t *MAWK, register CELL *target, register CELL *source)
{
	switch (target->type = source->type) {
	case C_NOINIT:
	case C_SPACE:
	case C_SNULL:
		break;

	case C_NUM:
		target->dval = source->dval;
		break;

	case C_STRNUM:
		target->dval = source->dval;
		/* fall thru */

	case C_REPL:
	case C_MBSTRN:
	case C_STRING:
		string(source)->ref_cnt++;
		/* fall thru */

	case C_RE:
		target->ptr = source->ptr;
		break;

	case C_REPLV:
		mawk_replv_cpy(MAWK, target, source);
		break;

	default:
		mawk_bozo(MAWK, "bad cell passed to mawk_cellcpy()");
		break;
	}
	return target;
}

#ifdef	 DEBUG
void DB_cell_destroy(mawk_state_t *MAWK, register CELL *cp)	/* HANGOVER time */
{
	switch (cp->type) {
	case C_NOINIT:
	case C_NUM:
		break;

	case C_MBSTRN:
	case C_STRING:
	case C_STRNUM:
		if (--string(cp)->ref_cnt == 0)
			zfree(MAWK, string(cp), string(cp)->len + STRING_OH);
		break;

	case C_RE:
		mawk_bozo(MAWK, "cell destroy called on RE cell");
	default:
		mawk_bozo(MAWK, "cell destroy called on bad cell type");
	}
}

#endif



/* convert a number d to a field index	$d -> $i */
static int d_to_index(mawk_state_t *MAWK, mawk_num_t d)
{

	if (d > MAX_FIELD)
		mawk_rt_overflow(MAWK, "maximum number of fields", MAX_FIELD);

	if (d >= NUM_ZERO)
		return (int) d;

	/* might include nan */
	mawk_rt_error(MAWK, "negative field index $%.6g", d);
	return 0;											/* shutup */
}
