
/********************************************
mawk_hash.c

libmawk changes (C) 2009-2010, Tibor 'Igor2' Palinkas;
based on mawk code coming with the below copyright:

copyright 1991, Michael D. Brennan

This is a source file for mawk, an implementation of
the AWK programming language.

Mawk is distributed without warranty under the terms of
the GNU General Public License, version 2, 1991.
********************************************/

#include "mawk.h"
#include "memory.h"
#include "symtype.h"
#include <string.h>

unsigned mawk_hash(register const char *s)
{
	register unsigned h = 0;

	while (*s)
		h += h + *s++;
	return h;
}

static HASHNODE *delete(mawk_state_t *, char *);

/*
mawk_insert a string in the symbol table.
Caller knows the symbol is not there
-- used for initialization
*/

SYMTAB *mawk_insert(mawk_state_t *MAWK, char *s)
{
	register HASHNODE *p = ZMALLOC(HASHNODE);
	register unsigned h;

	p->link = MAWK->hash_table[h = mawk_hash(s) % HASH_PRIME];
	p->symtab.name = s;
	MAWK->hash_table[h] = p;
	return &p->symtab;
}

/* Find s in the symbol table,
   if not there and alloc !=0 mawk_insert it (s must be dup'ed)  */

SYMTAB *mawk_find(mawk_state_t *MAWK, const char *s, int alloc)
{
	register HASHNODE *p;
	HASHNODE *q;
	unsigned h;

	p = MAWK->hash_table[h = mawk_hash(s) % HASH_PRIME];
	q = (HASHNODE *) 0;
	while (1) {
		if (!p) {
			if (alloc) {
				p = ZMALLOC(HASHNODE);
				p->symtab.type = ST_NONE;
				p->symtab.name = strcpy(zmalloc(MAWK, strlen(s) + 1), s);
				break;
			}
			else
				return NULL;
		}

		if (strcmp(p->symtab.name, s) == 0) {	/* found */
			if (!q)										/* already at the front */
				return &p->symtab;
			else {										/* delete from the list */

				q->link = p->link;
				break;
			}
		}

		q = p;
		p = p->link;
	}
	/* put p on front of the list */
	p->link = MAWK->hash_table[h];
	MAWK->hash_table[h] = p;
	return &p->symtab;
}


/* remove a node from the mawk_hash table
   return a ptr to the node */

static HASHNODE *delete(mawk_state_t *MAWK, char *s)
{
	register HASHNODE *p;
	HASHNODE *q = (HASHNODE *) 0;
	unsigned h;

	p = MAWK->hash_table[MAWK->last_hash = h = mawk_hash(s) % HASH_PRIME];
	while (p) {
		if (strcmp(p->symtab.name, s) == 0) {	/* found */
			if (q)
				q->link = p->link;
			else
				MAWK->hash_table[h] = p->link;
			return p;
		}
		else {
			q = p;
			p = p->link;
		}
	}

#ifdef	DEBUG										/* we should not ever get here */
	mawk_bozo(MAWK, "delete");
#endif
	return (HASHNODE *) 0;
}

/* store a global id on the save list,
   return a ptr to the local symtab  */
SYMTAB *mawk_save_id(mawk_state_t *MAWK, char *s)
{
	HASHNODE *p, *q;
	unsigned h;

	p = delete(MAWK, s);
	q = ZMALLOC(HASHNODE);
	q->symtab.type = ST_LOCAL_NONE;
	q->symtab.name = p->symtab.name;
	/* put q in the mawk_hash table */
	q->link = MAWK->hash_table[h = MAWK->last_hash];
	MAWK->hash_table[h] = q;

	/* save p */
	p->link = MAWK->save_list;
	MAWK->save_list = p;

	return &q->symtab;
}

/* restore all global indentifiers */
void mawk_restore_ids(mawk_state_t * MAWK)
{
	register HASHNODE *p, *q;
	register unsigned h;

	q = MAWK->save_list;
	MAWK->save_list = (HASHNODE *) 0;
	while (q) {
		p = q;
		q = q->link;
		zfree(MAWK, delete(MAWK, p->symtab.name), sizeof(HASHNODE));
		p->link = MAWK->hash_table[h = MAWK->last_hash];
		MAWK->hash_table[h] = p;
	}
}


/* search the symbol table backwards for the
   disassembler.  This is slow -- so what
*/

const char *mawk_reverse_uk = "unknown";

const char *mawk_reverse_find(mawk_state_t *MAWK, int type, PTR ptr)
{
	CELL *cp;
	ARRAY array;

	int i;
	HASHNODE *p;


	switch (type) {
	case ST_VAR:
	case ST_FIELD:
		cp = *(CELL **) ptr;
		break;

	case ST_ARRAY:
		array = *(ARRAY *) ptr;
		break;

	default:
		return mawk_reverse_uk;
	}

	for (i = 0; i < HASH_PRIME; i++) {
		p = MAWK->hash_table[i];
		while (p) {
			if (p->symtab.type == type) {
				switch (type) {
				case ST_VAR:
				case ST_FIELD:
					if (cp == p->symtab.stval.cp)
						return p->symtab.name;
					break;

				case ST_ARRAY:
					if (array == p->symtab.stval.array)
						return p->symtab.name;
					break;
				}
			}

			p = p->link;
		}
	}
	return mawk_reverse_uk;
}

/* free global hash entries */
void mawk_hash_clear(mawk_state_t *MAWK)
{
	register HASHNODE *p, *next;
	int n;

	for(n = 0; n < HASH_PRIME; n++) {
		for(p = MAWK->hash_table[n]; p != NULL; p = next) {
			next = p->link;
			if (p->symtab.type == ST_VAR)
				cell_destroy(p->symtab.stval.cp);
			else if (p->symtab.type == ST_ARRAY)
				mawk_array_clear(MAWK, p->symtab.stval.array);
		}
	}
}
