
/********************************************
mawk.h

libmawk changes (C) 2009-2012, Tibor 'Igor2' Palinkas;
based on mawk code coming with the below copyright:

copyright 1991-94, Michael D. Brennan

This is a source file for mawk, an implementation of
the AWK programming language.

Mawk is distributed without warranty under the terms of
the GNU General Public License, version 2, 1991.
********************************************/

#ifndef  MAWK_H
#define  MAWK_H

#include <libmawk/conf.h>
#include <stdlib.h>
#include <libmawk/nstd.h>
#include <stdio.h>

typedef enum mawk_errno_e {
	MAWK_ESUCCES    = 0,
	MAWK_ECANTOPEN  = -1,
	MAWK_EHDRSIZE   = -2,
	MAWK_EFILEMAGIC = -3,
	MAWK_EBYTEORDER = -4,
	MAWK_EVERSION   = -5,
	MAWK_EINSTSIZE  = -6,
	MAWK_ENUMSIZE   = -7,
	MAWK_EALLOC     = -8,
	MAWK_EWRONGVAL  = -9,

	MAWK_Elast      = -9
} mawk_errno_t;
const char *mawk_strerror(mawk_errno_t err);

typedef struct mawk_state_s mawk_state_t;
typedef struct fcall  FCALL_REC;

#include <libmawk/types.h>
#include <setjmp.h>

#define  NUM_PFIELDS		5

#define  SPRINTF_SZ	sizeof(MAWK->tempbuff)

typedef struct pfile {
	struct pfile *link;
	const char *fname;
	char bytecode; /* 1 if file is expected to be bytecode */
} PFILE;

typedef struct {
	int fd;
	FILE *fp;											/* NULL unless interactive */
	char *buff;
	char *buffp;
	unsigned nbuffs;							/* sizeof *buff in BUFFSZs */
	int flags;

	/* artificial buff can inject characters in a stream - this how we bypass file handling when the caller wants the script to run on memory buffers */
	char *artificial_buff;				/* must be NULL or malloc()'d */
	int artificial_size;					/* number of input characters left in  artificial_buff */
} FIN;

typedef struct array {
	PTR ptr;											/* What this points to depends on the type */
	unsigned size;								/* number of elts in the table */
	unsigned limit;								/* Meaning depends on type */
	unsigned hmask;								/* bitwise and with hash value to get table index */
	short type;										/* values in AY_NULL .. AY_SPLIT */
} *ARRAY;

typedef struct {
	INST *base, *limit, *warn, *ptr;
} CODEBLOCK;

/*------------------------
  user defined functions
  ------------------------*/

typedef struct fblock {
	char *name;
	INST *code;
	unsigned short nargs;
	char *typev;									/* array of size nargs holding types */
} FBLOCK;												/* function block */

typedef struct jmp {
	struct jmp *link;
	int source_offset;
} JMP;

typedef struct bc {
	struct bc *link;							/* stack as linked list */
	int type;											/* 'B' or 'C' or mark start with 0 */
	int source_offset;						/* position of _JMP  */
} BC;

/* a stack to hold some pieces of code while
   reorganizing loops .
*/

typedef struct mc {							/* mc -- move code */
	struct mc *link;
	INST *code;										/* the save code */
	unsigned len;									/* its length */
	int scope;										/* its scope */
	int move_level;								/* size of this stack when coded */
	FBLOCK *fbp;									/* if scope FUNCT */
	int offset;										/* distance from its code base */
} MC;

struct child {
	int pid;
	int exit_status;
	struct child *link;
};

#define	 SAFETY	   16
#define	 DANGER	   (EVAL_STACK_SIZE-SAFETY)

#define	 ET_END	    9
typedef struct {
	char in, out;
} mawk_escape_t;

struct mawk_fdump {
	struct mawk_fdump *link;
	FBLOCK *fbp;
};

/* We store dynamically created files on a linked linear
   list with move to the front (big surprise)  */

typedef struct file {
	struct file *link;
	STRING *name;
	short type;
	int pid;											/* we need to wait() when we close a pipe */

	PTR ptr;											/* FIN*   or  FILE*   */
} FILE_NODE;

typedef struct {
	char type;
	char c;
	PTR ptr;											/* STRING* or RE machine* */
} SEPARATOR;

/* struct to hold info about builtins */
typedef struct {
	char *name;
	PF_CP fp;											/* ptr to function that does the builtin */
	unsigned char min_args, max_args;
/* info for parser to check correct number of arguments */
} BI_REC;

typedef struct {
	char *name;
	char type;
	unsigned char offset;					/* offset in stack frame for local vars */
	union {
		CELL *cp;
		int kw;
		PF_CP fp;
		const BI_REC *bip;
		ARRAY array;
		FBLOCK *fbp;
		struct {
			CELL *(*callback) (mawk_state_t * context, CELL * sp, int a_args);
			void *userdata;
		} c_function;
	} stval;
} SYMTAB;

typedef struct hash {
	struct hash *link;
	SYMTAB symtab;
} HASHNODE;

#define POOLSZ	    16
#define ZBLOCKSZ    8
#define ZSHIFT      3
typedef union zblock {
	char dummy[ZBLOCKSZ];
	union zblock *link;
} ZBLOCK;

/* ZBLOCKS of sizes 1, 2, ... 16
   which is bytes of sizes 8, 16, ... , 128
   are stored on the linked linear lists in
   pool[0], pool[1], ... , pool[15]
*/

typedef struct re_node {
	STRING *sval;
	PTR re;
	struct re_node *link;
} RE_NODE;

typedef struct repl_node {
	struct repl_node *link;
	STRING *sval;									/* the input */
	CELL *cp;											/* the output */
} REPL_NODE;

typedef struct spov {
	struct spov *link;
	STRING *sval;
} SPLIT_OV;

/* ---------------------------------------------------------------------- */

typedef struct mawk_debug_callstack_s mawk_debug_callstack_t;

struct mawk_debug_callstack_s {
	FBLOCK *f;
	mawk_debug_callstack_t *next;
};

typedef struct mawk_parse_state_s {
	const char *pfile_name;				/* program input file */
	char pfile_bytecode; /* 1 if program input file is expected to be bytecode */
	int code_move_level;					/* used to as one part of unique identification of context when moving code.  Global for communication with parser. */
	STRING *program_string;
	unsigned char *buffer;
	unsigned char *buffp;
	/* unsigned so it works with 8 bit chars */
	int program_fd;
	int eof_flag;
} mawk_parse_state_t;


typedef struct mawk_mm_s mawk_mm_t;

struct mawk_mm_s {
	mawk_mm_t *prev, *next;
	int size;
	char data[1];                     /* actual data */
} ;

struct mawk_state_s {

#ifdef   DEBUG
#define  YYDEBUG  1
	int yydebug;									/* print parse if on */
	int dump_RE;
#endif

	short posix_space_flag, interactive_flag;

/* a well known string */
	STRING null_str;

/* a useful scratch area */
	union {
		STRING *_split_buff[MAX_SPLIT];
		char _string_buff[MIN_SPRINTF];
	} tempbuff;


	/* help with casts */
	int *mpow2;

	CELL field[FBANK_SZ + NUM_PFIELDS];
	/* $0, $1 ... $(MAX_SPLIT), NF, RS, RS, CONVFMT, OFMT */

	/* more fields if needed go here */
	CELL *fbank[NUM_FBANK];				/* fbank[0] == field */


	/* these are used by the parser, scanner and error messages
	   from the compile  */

	PFILE *pfile_list, *pfile_list_tail;

	int current_token;
	unsigned token_lineno;				/* lineno of current token */
	unsigned compile_error_count;
	int paren_cnt, brace_cnt;
	int print_flag, getline_flag;
	short mawk_state;
	char *progname;								/* for error messages */
	unsigned rt_nr, rt_fnr;				/* ditto */

	/* this can be moved and enlarged  by -W sprintf=num  */
	char *sprintf_buff;
	char *sprintf_limit;

	FIN *main_fin;
	ARRAY Argv;										/* to the user this is ARGV  */
	mawk_num_t argi;										/* index of next ARGV[argi] to try to open */
	unsigned lineno;
	int NR_flag;									/* are we tracking NR */

	CODEBLOCK active_code;
	CODEBLOCK *main_code_p, *begin_code_p, *end_code_p;
	INST *begin_start, *main_start, *end_start;
	unsigned begin_size, main_size;
	INST *execution_start;

	int dump_code_flag;						/* if on dump internal code */

	INST *restart_label;					/* control flow labels */
	INST *next_label;
	CELL tc;											/*useful temp */
	int scope;
	FBLOCK *active_funct;					/* when scope is SCOPE_FUNCT  */
	JMP *jmp_top;
	BC *bc_top;
	MC *mc_top;

	mawk_parse_state_t ps;				/* current parse state */
	ARRAY scripts_loaded;					/* a hash indeced by full paths of scripts already loaded */
	mawk_parse_state_t *mawk_parser_stack;	/* parse state stack for "include" */
	int pstack_alloced, pstack_used;

	int check_progress;						/* flag that indicates call_arg_check() was able to type check some call arguments */
	struct child *child_list;			/* dead children are kept on this list */
	unsigned repl_cnt;						/* number of global replacements */
	long seed;										/* must be >=1 and < 2^31-1 */
	CELL cseed;										/* argument of last call to srand() */
	CELL eval_stack[EVAL_STACK_SIZE];
	CELL *stack_base;							/* these can move for deep recursion */
	CELL *stack_danger;
	int exit_code;
#	ifdef   HAVE_STRTOD_OVF_BUG
	fp_except entry_mask;
	fp_except working_mask;
#	endif
	mawk_escape_t escape_test[ET_END + 1];
	struct mawk_fdump *fdump_list;	/* linked list of all user functions */
	FILE_NODE *file_list;
	char *shell;									/* hardwire to /bin/sh for portability of programs */

	int max_field;								/* maximum field actually created */
/* a description of how to split based on RS.
   If RS is changed, so is rs_shadow */
	SEPARATOR rs_shadow;
	/* a splitting CELL version of FS */
	CELL fs_shadow;
	int nf;												/* nf holds the true value of NF.  If nf < 0 , then NF has not been computed, i.e., $0 has not been split */
	HASHNODE *hash_table[HASH_PRIME];
	HASHNODE *save_list;					/* when processing user functions,  global ids which are replaced by local ids are saved on this list */
	unsigned last_hash;

	/* large block allocator (memory accounting and free-later mechanism) */
	mawk_mm_t *mawk_mm_head;
	int mm_used, mm_max;

	/* small block allocator in zmalloc.[ch] (pooling) */
	ZBLOCK *pool[POOLSZ];         /* pool of blocks already free'd, indexed by size (in blocks) */
	unsigned amt_avail;           /* how many blocks are unclaimed at the end of ->avail */
	ZBLOCK *avail;                /* the chunk we split up for new allocations */

	RE_NODE *re_list;							/* a list of compiled regular expressions */
	REPL_NODE *repl_list;					/* here's our old friend linked linear list with move to the front for compilation of replacement CELLs   */
	char scan_code[256];
	SPLIT_OV *split_ov_list;

	void *userdata;								/* during calls to C functions, userdata has the value that it had during registration of that function */

	int last_token_lineno;				/* last token line number to detect source line change for adding debug info */
	mawk_debug_callstack_t *debug_callstack;

	FCALL_REC *resolve_list;
	void *lvalp;
	CELL code_call_id_dummy;

	/* should be a bitfield! */
	int debug_symbols;						/* add location infoand other debug symbol data to the code */
	int separate_begin;						/* if not zero, after running BEGIN blocks, no main block is automaticly executed */
	int dont_close_std;						/* if not zero, do not close stdout and stderr */
	int suppress_undefined_function_warning;	/* if not zero, do not warn about functions undefined */

	int do_exit;									/* non-zero if we should exit immediately (added for exiting from the parser) */
	int binary_loaded;            /* non-zero if no text parsing is required (binary file has been loaded) */
};

#define EXECUTION       1				/* other state is 0 compiling */

	/* anonymous union */
#define  string_buff	MAWK->tempbuff._string_buff
#define  split_buff	MAWK->tempbuff._split_buff



/* macro to test the type of two adjacent cells */
#define TEST2(cp)  (MAWK->mpow2[(cp)->type]+MAWK->mpow2[((cp)+1)->type])

/* macro to get at the string part of a CELL */
#define string(cp) ((STRING *)(cp)->ptr)

#ifdef   DEBUG
#define cell_destroy(cp)  DB_cell_destroy(MAWK, cp)
#else

#define cell_destroy(cp)   if ( (cp)->type >= C_STRING &&\
                           -- string(cp)->ref_cnt == 0 )\
                        zfree(MAWK, string(cp),string(cp)->len+STRING_OH);else
#endif

/*  prototypes  */

void mawk_cast1_to_s(mawk_state_t *, CELL *);
void mawk_cast1_to_d(mawk_state_t *, CELL *);
void mawk_cast2_to_s(mawk_state_t *, CELL *);
void mawk_cast2_to_d(mawk_state_t *, CELL *);
void mawk_cast_to_RE(mawk_state_t *, CELL *);
void mawk_cast_for_split(mawk_state_t *, CELL *);
void mawk_check_strnum(mawk_state_t *, CELL *);
void mawk_cast_to_REPL(mawk_state_t *, CELL *);

#define d_to_i(d)     ((int)mawk_d_to_I(d))


int test(mawk_state_t *, CELL *);	/* test for null non-null */
CELL *mawk_cellcpy(mawk_state_t *, CELL *, CELL *);
CELL *repl_cpy(CELL *, CELL *);
void DB_cell_destroy(mawk_state_t *, CELL *);
void overflow(mawk_state_t *, char *, unsigned);
void mawk_rt_overflow(mawk_state_t * MAWK, char *, unsigned);
void mawk_rt_error(mawk_state_t *, const char *, ...);
void mawk_set_errno(mawk_state_t * MAWK, const char *error);


void mawk_exit_(mawk_state_t *, int);
#define mawk_exitval(MAWK, x, RETVAL) \
	do { \
		mawk_exit_(MAWK, x); \
		return RETVAL; \
	} while(0);

#define mawk_exit(MAWK, x) \
	do { \
		mawk_exit_(MAWK, x); \
		return; \
	} while(0);

void mawk_da(mawk_state_t *, INST *, FILE *);
char *mawk_str_str(char *, char *, unsigned);
char *mawk_rm_escape(mawk_state_t *, char *);
char *mawk_re_pos_match(mawk_state_t *, char *, PTR, unsigned *);
int mawk_binmode(void);


void mawk_bozo(mawk_state_t *, char *);
void mawk_errmsg(mawk_state_t *, int, char *, ...);
void mawk_compile_error(mawk_state_t *, const char *, ...);

void mawk_execute(mawk_state_t *, INST *, CELL *, CELL *);
const char *mawk_find_kw_str(int);

CELL *mawk_call(mawk_state_t * MAWK, CELL * sp, FBLOCK * fbp, int a_args);
void mawk_overflow(mawk_state_t * MAWK, char *s, unsigned size);
int mawk_test(mawk_state_t *MAWK, register CELL *cp);
void mawk_bi_vars_init(mawk_state_t * MAWK);
void mawk_bi_funct_init(mawk_state_t * MAWK);
void mawk_code_init(mawk_state_t *MAWK);
void mawk_parse(mawk_state_t *);

#ifndef MAWK_NO_FLOAT
#	ifdef HAVE_STRTOD_OVF_BUG
		double strtod_with_ovf_bug(const char *, char **);
#		define strtod  strtod_with_ovf_bug
#	endif
#endif

#endif /* MAWK_H */
